"""Models and data structures used by the library"""

import enum
import functools
import logging
import typing
import uuid
from collections.abc import Awaitable, Callable, Iterable, Mapping

from attrs import define, evolve, field

logger = logging.getLogger(__name__)


class DeviceType(enum.Enum):
    """Enumeration indicating the type of the device"""

    GATEWAY = "gateway"
    HVAC = "hvac"


class SystemMode(enum.Enum):
    """HVAC system mode"""

    OFF = 0
    HEAT = 4


@define
class DeviceAttributes:
    """Common device attributes"""

    id: uuid.UUID
    type: DeviceType
    device_code: str
    model: str
    name: str


@define
class Occupant:
    """Occupant of a managed device"""

    id: uuid.UUID
    identity_id: str


@define
class GatewayAttributes(DeviceAttributes):
    """Gateway attributes"""

    type: typing.Literal[DeviceType.GATEWAY]
    occupant: Occupant


@define
class HvacAttributes(DeviceAttributes):
    """Attributes of a HVAC device"""

    type: typing.Literal[DeviceType.HVAC]

    system_mode: SystemMode | None = field(default=None)
    """The state of the device"""

    temperature: float | None = field(default=None)
    """The setpoint temperature"""

    current_temperature: float | None = field(default=None)
    """The current temperature as measured by the device"""

    min_temp: float | None = field(default=None)
    """Minimum setpoint temperature"""

    max_temp: float | None = field(default=None)
    """Maximum setpoint temperature"""


DeviceChangeSubscriber = Callable[["Device", Mapping[str, typing.Any]], None]


AttributesType = typing.TypeVar(  # pylint: disable=invalid-name
    "AttributesType", bound=DeviceAttributes
)
"""Type variable"""


class Device(typing.Generic[AttributesType]):
    """A managed device

    A device object is a dynamic handle for the device data managed by a client.
    It responds to direct and indirect state changes both from the client and
    other sources.  The instantaneous state of the device can be retrieved using
    the :meth:`get_attributes()` method.

    Arguments:
      attributes: the initial attributes
      dispatch_refresh: dispatch request to refresh the state of this device
      dispatch_update: dispatch request to update the state of this device
    """

    def __init__(
        self,
        attributes: AttributesType,
        dispatch_refresh: Callable[["Device"], Awaitable[None]],
        dispatch_update: Callable[
            ["Device", Mapping[str, typing.Any]], Awaitable[None]
        ],
    ):
        self._attributes = attributes
        self._dispatch_refresh = dispatch_refresh
        self._dispatch_update = dispatch_update
        self._subscribers: list[DeviceChangeSubscriber] = []

    def get_attributes(self) -> AttributesType:
        """Get the attributes associated with the device"""
        return self._attributes

    def get_id(self) -> uuid.UUID:
        """Get device id"""
        return self._attributes.id

    def get_type(self) -> DeviceType:
        """Get device type"""
        return self._attributes.type

    def get_device_code(self) -> str:
        """Get device code"""
        return self._attributes.device_code

    def get_model(self) -> str:
        """Get device code"""
        return self._attributes.model

    def get_name(self) -> str:
        """Get device code"""
        return self._attributes.name

    def set_attributes(self, changes: Mapping[str, typing.Any]):
        """Set new value for attributes

        Note that this method only changes the attributes in-memory, and the
        changes are not dispatched to the cloud service.

        Arguments:
          changes: dictionary containing new values for the attributes
        """
        if changes:
            self._attributes = evolve(self._attributes, **changes)
            for subscriber in self._subscribers:
                try:
                    subscriber(self, changes)
                except Exception:  # pylint: disable=broad-exception-caught
                    logger.exception(
                        "Failed to invoke subscriber callback for device %r",
                        self.get_attributes(),
                    )

    async def refresh(self):
        """Refresh the state of the device

        This call will delegate to :meth:`Client.refresh_device_state()`
        """
        await self._dispatch_refresh(self)

    async def update(self, changes: Mapping[str, typing.Any]):
        """Update the state of the device

        This call will delegate to :meth:`Client.update_device_state()`
        """
        await self._dispatch_update(self, changes)

    def subscribe(self, callback: DeviceChangeSubscriber):
        """Register a callback to be notified when the state of a device changes

        Arguments:
          callback: the callback that will be called with the updated device and
                    dictionary of the changed attributes
        """
        self._subscribers.append(callback)

    def unsubscribe(self, callback: DeviceChangeSubscriber):
        """Remove a callback previously registered with :meth:`subscribe()`

        Arguments:
          callback: the callback to be removed from subscribers
        """
        self._subscribers.remove(callback)


class HvacDevice(Device[HvacAttributes]):
    """A HVAC device (smart thermostat)"""

    def get_system_mode(self) -> SystemMode | None:
        """Get state of the device"""
        return self._attributes.system_mode

    def get_temperature(self) -> float | None:
        """Get the setpoint temperature"""
        return self._attributes.temperature

    def get_current_temperature(self) -> float | None:
        """Get the current temperature as measured by the device"""
        return self._attributes.current_temperature

    def get_min_temp(self) -> float | None:
        """Get the minimum setpoint temperature"""
        return self._attributes.min_temp

    def get_max_temp(self) -> float | None:
        """Get the maximum setpoint temperature"""
        return self._attributes.max_temp

    async def update_system_mode(self, system_mode: SystemMode):
        """Update the system mode"""
        await self.update({"system_mode": system_mode})

    async def update_temperature(self, temperature: float):
        """Update the setpoint temperature"""
        await self.update({"temperature": temperature})


class Gateway(Device[GatewayAttributes]):
    """A gateway acting between Unisenza IoT devices and the cloud service

    Arguments:
      attributes: the gateway attributes
      children: initial attributes of the children managed by the gateway
      dispatch_refresh: dispatch request to refresh the state of this device
      dispatch_update: dispatch request to update the state of this device
    """

    def __init__(
        self,
        attributes: GatewayAttributes,
        children: Iterable[HvacAttributes],
        dispatch_refresh: Callable[["Gateway", Device], Awaitable[None]],
        dispatch_update: Callable[
            ["Gateway", Device, Mapping[str, typing.Any]], Awaitable[None]
        ],
    ):
        bound_dispatch_refresh = functools.partial(dispatch_refresh, self)
        bound_dispatch_update = functools.partial(dispatch_update, self)
        super().__init__(attributes, bound_dispatch_refresh, bound_dispatch_update)
        self._children = [
            HvacDevice(child, bound_dispatch_refresh, bound_dispatch_update)
            for child in children
        ]

    def get_children(self) -> list[HvacDevice]:
        """Get the children of this device"""
        return self._children

    def get_occupant(self) -> Occupant:
        """Get the occupant of the gateway"""
        return self._attributes.occupant



"""Entry point to the command-line interface"""


def main():
    """Run ``pyupgw`` as script"""

    try:
        from .cli import cli  # pylint: disable=import-outside-toplevel
    except ImportError:
        print("Unable to load CLI. Perhaps missing [cli] extras?")
    else:
        cli()  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    main()



"""Interactive text-user interface"""

import asyncio
import contextlib
import curses
import signal
import sys
import typing

from blessed import Terminal
from rich import print
from rich.console import Console, Group
from rich.live import Live
from rich.panel import Panel
from rich.table import Table

from pyupgw import Client, Device, HvacDevice, SystemMode, create_client

HELP_PANEL = Panel(
    "(:arrow_up::arrow_down:) Select device, (O)ff, (H)eat, (+-) Adjust temperature, (Enter) Confirm"
)


# pylint: disable=too-many-instance-attributes
class Application(contextlib.AbstractAsyncContextManager):
    """Application state"""

    def __init__(self, client: Client, term: Terminal):
        self._client = client
        self._term = term
        self._queue: asyncio.Queue[tuple[Device, typing.Any]] = asyncio.Queue()
        self._exit_event = asyncio.Event()
        self._device_index = 0
        self._inkey_task = self._create_inkey_task()
        self._get_change_task = self._create_get_key_task()
        self._exit_task = self._create_exit_task()
        self._tasks: list[asyncio.Task] = [
            self._inkey_task,
            self._get_change_task,
            self._exit_task,
        ]
        self._devices: list[HvacDevice] = []
        self._pending_temperature: float | None = None

    def _enqueue_change(self, device: Device, changes: typing.Any):
        self._queue.put_nowait((device, changes))

    async def __aenter__(self):
        await self._client.refresh_all_devices()
        for _, device in self._client.get_devices():
            device.subscribe(self._enqueue_change)
        self._devices = list(device for _, device in self._client.get_devices())
        loop = asyncio.get_event_loop()
        loop.add_signal_handler(signal.SIGINT, self._exit_event.set)
        loop.add_signal_handler(signal.SIGTERM, self._exit_event.set)
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        for task in self._tasks:
            task.cancel()

    async def tick(self, console: Console):
        """Handle application events"""
        done, pending = await asyncio.wait(
            self._tasks,
            return_when=asyncio.FIRST_COMPLETED,
        )
        self._tasks = list(pending)
        if self._inkey_task in done:
            self._handle_inkey_task()
        if self._get_change_task in done:
            self._handle_get_change_task(console)

    def get_renderable(self):
        """Get application renderable"""
        table = Table(title="Devices", expand=True)
        table.add_column("Device")
        table.add_column("Gateway")
        table.add_column("State")
        table.add_column("Temperature")
        table.add_column("Current temperature")

        for current, (gateway, device) in enumerate(self._client.get_devices()):
            table.add_row(
                f"{':arrow_forward:' if self._device_index == current else ' '} {device.get_name()}",
                gateway.get_name(),
                str(
                    system_mode.name
                    if (system_mode := device.get_system_mode())
                    else None
                ),
                self._temperature_renderable(current, device),
                str(device.get_current_temperature()),
            )

        return Group(table, HELP_PANEL)

    def done(self):
        """Return True if done, False otherwise"""
        return self._exit_event.is_set()

    def _temperature_renderable(self, current: int, device: HvacDevice):
        if current == self._device_index and self._pending_temperature is not None:
            return f"[blink]{self._pending_temperature}"
        return str(device.get_temperature())

    def _create_inkey_task(self):
        return asyncio.create_task(asyncio.to_thread(self._term.inkey))

    def _create_get_key_task(self):
        return asyncio.create_task(self._queue.get())

    def _create_exit_task(self):
        return asyncio.create_task(self._exit_event.wait())

    def _current_device(self):
        return self._devices[self._device_index]

    def _handle_inkey_task(self):
        assert self._inkey_task.done()
        key = self._inkey_task.result()
        if key.code == curses.KEY_UP:
            self._pending_temperature = None
            self._device_index = max(self._device_index - 1, 0)
        elif key.code == curses.KEY_DOWN:
            self._pending_temperature = None
            self._device_index = min(self._device_index + 1, len(self._devices) - 1)
        elif key.upper() == "O" and self._devices:
            self._pending_temperature = None
            self._tasks.append(
                asyncio.create_task(
                    self._current_device().update_system_mode(SystemMode.OFF),
                )
            )
        elif key.upper() == "H" and self._devices:
            self._pending_temperature = None
            self._tasks.append(
                asyncio.create_task(
                    self._current_device().update_system_mode(SystemMode.HEAT),
                )
            )
        elif key.code == curses.KEY_ENTER and self._pending_temperature is not None:
            self._tasks.append(
                asyncio.create_task(
                    self._current_device().update_temperature(self._pending_temperature)
                )
            )
            self._pending_temperature = None
        elif key == "+":
            current_device = self._current_device()
            if self._pending_temperature is None:
                self._pending_temperature = current_device.get_temperature()
            self._pending_temperature = min(
                self._pending_temperature + 0.5, current_device.get_max_temp()
            )
        elif key == "-":
            current_device = self._current_device()
            if self._pending_temperature is None:
                self._pending_temperature = current_device.get_temperature()
            self._pending_temperature = max(
                self._pending_temperature - 0.5, current_device.get_min_temp()
            )
        self._inkey_task = self._create_inkey_task()
        self._tasks.append(self._inkey_task)

    def _handle_get_change_task(self, console: Console):
        assert self._get_change_task.done()
        changed_device, changes = self._get_change_task.result()
        console.log(f"{changed_device.get_name()} changed:", changes)
        self._get_change_task = self._create_get_key_task()
        self._tasks.append(self._get_change_task)


async def _tui_main(term: Terminal, username: str, password: str):
    async with create_client(username, password) as client:
        async with Application(client, term) as app:
            with Live(app.get_renderable()) as live:
                while not app.done():
                    await app.tick(live.console)
                    live.update(app.get_renderable(), refresh=True)


def tui(username: str, password: str):
    """Text-user interface"""

    term = Terminal()
    if not term.is_a_tty:
        print("[b red]Interactive mode supported only in TTY", file=sys.stderr)
        sys.exit(1)

    with term.fullscreen(), term.cbreak():
        asyncio.run(_tui_main(term, username, password))



"""Command line interface"""

# pylint: disable=wrong-import-position,wrong-import-order,redefined-builtin

from dotenv import load_dotenv

load_dotenv()

import asyncio
import functools
import os
import sys
import typing
from collections.abc import Iterable

import rich_click as click
from rich.console import Console
from rich.live import Live
from rich.table import Table

from pyupgw import Client, Device, Gateway, HvacDevice, SystemMode, create_client

from .commands import get as get_impl
from .commands import list_devices as list_impl
from .commands import update as update_impl
from .tui import tui


@click.group(invoke_without_command=True)
@click.option(
    "-u",
    "--username",
    help="Unisenza Plus user. Alternatively can be loaded from the PYUPGW_USERNAME environment variable.",
)
@click.option(
    "-p",
    "--password",
    help="Password for the user. Alternatively can be loaded from the PYUPGW_PASSWORD environment variable.",
)
@click.pass_context
def cli(ctx: click.Context, username: str | None, password: str | None):
    """Command-line interface for Unisenza Plus

    When invoked without subcommand, starts the client in interactive mode.

    The author of this tool is not affiliated with Purmo, the vendor of Unisenza
    Plus, in any way.  The tool and its conformance to the Unisenza Plus API is
    implemented on a best-effort basis.  No warranty of any kind is provided.
    """

    username = username or os.getenv("PYUPGW_USERNAME")
    password = password or os.getenv("PYUPGW_PASSWORD")

    if not username or not password:
        click.echo("Username and password required")
        sys.exit(0)

    ctx.obj = {
        "username": username,
        "password": password,
    }

    if ctx.invoked_subcommand is None:
        tui(username, password)


def with_client(func):
    """Decorate for a command that gets client as first argument"""

    @functools.wraps(func)
    @click.pass_context
    def inner(ctx: click.Context, *args, **kwargs):
        async def _inner_async(*args, **kwargs):
            async with create_client(**ctx.obj) as client:
                return await func(client, *args, **kwargs)

        return asyncio.run(_inner_async(*args, **kwargs))

    return inner


@cli.command()
@with_client
async def list(client: Client):
    """Print list of gateways and devices with details"""

    return await list_impl(client)


@cli.command()
@click.argument("device", required=True)
@with_client
async def get(client: Client, device: str):
    """Print details of a single device

    The DEVICE can be identified either by id, name or device code."""

    return await get_impl(client, device)


@cli.command()
@click.argument("device", required=True)
@click.option("--temperature", "-t", type=float, help="New temperature setpoint")
@click.option(
    "--system-mode",
    "-s",
    type=click.Choice([sm.name for sm in SystemMode]),
    help="New system mode",
)
@with_client
async def update(
    client: Client, device: str, temperature: float | None, system_mode: str | None
):
    """Update the state of a device

    The DEVICE can be identified either by id, name or device code."""

    system_mode_enum = SystemMode[system_mode] if system_mode is not None else None

    return await update_impl(client, device, temperature, system_mode_enum)



"""Non-interactive commands"""

import sys
import typing

from rich import print
from rich.console import Group
from rich.pretty import Pretty
from rich.tree import Tree

from pyupgw import Client, Device, DeviceType, SystemMode


def _find_device_by_str(client: Client, needle: str):
    for _, device in client.get_devices():
        if needle in [
            device.get_name(),
            device.get_device_code(),
            str(device.get_id()),
        ]:
            return device
    return None


def _pretty_print_device(device: Device):
    if device.get_type() == DeviceType.GATEWAY:
        header = f":globe_with_meridians: [bold]Gateway: {device.get_name()}"
    else:
        header = f"Device: {device.get_name()}"
    return Group(header, Pretty(device.get_attributes()))


async def update(
    client: Client,
    device_needle: str,
    temperature: float | None,
    system_mode: SystemMode | None,
):
    """Update the state of a device"""

    device = _find_device_by_str(client, device_needle)
    if not device:
        print("[b red]Device not found", file=sys.stderr)
        sys.exit(1)

    changes: dict[str, typing.Any] = {}
    if temperature is not None:
        changes["temperature"] = temperature
    if system_mode is not None:
        changes["system_mode"] = system_mode
    await device.update(changes)


async def get(client: Client, device_needle: str):
    """Print details of a single device"""

    device = _find_device_by_str(client, device_needle)
    if not device:
        print("[b red]Device not found", file=sys.stderr)
        sys.exit(1)

    await device.refresh()
    print(_pretty_print_device(device))


async def list_devices(client: Client):
    """Print detailed view of managed gateways and devices"""
    await client.refresh_all_devices()
    tree = Tree("Device details", hide_root=True)
    for gateway in client.get_gateways():
        branch = tree.add(_pretty_print_device(gateway))
        for device in gateway.get_children():
            branch.add(_pretty_print_device(device))
    print(tree)



"""Python client for Unisenza Plus"""

__version__ = "0.3"

from .client import Client, ClientError, create_client
from .models import (
    Device,
    DeviceAttributes,
    DeviceType,
    Gateway,
    GatewayAttributes,
    HvacAttributes,
    HvacDevice,
    Occupant,
    SystemMode,
)



"""
Low level API

This module is meant for developing the high level client, and should rarely be
used directly.  Breaking changes may be introduced between minor versions.
"""

# pylint: disable=too-many-arguments

import asyncio
import contextlib
import os
import threading
import typing
import uuid
from collections.abc import Awaitable

import aiohttp
from awscrt.auth import AwsCredentialsProvider
from awscrt.io import ClientTlsContext, TlsContextOptions
from awsiot.iotshadow import IotShadowClient
from awsiot.mqtt_connection_builder import websockets_with_default_aws_signing
from pycognito import Cognito

PYUPGW_AWS_CLIENT_ID = os.getenv(
    "PYUPGW_AWS_CLIENT_ID", default="63qkc36u3eje4lp8ums9njmarv"
)
PYUPGW_AWS_REGION = os.getenv("PYUPGW_AWS_REGION", default="eu-central-1")
PYUPGW_AWS_USER_POOL_ID = os.getenv(
    "PYUPGW_AWS_USER_POOL_ID", default="eu-central-1_HfciXliKM"
)
PYUPGW_AWS_ID_PROVIDER = os.getenv(
    "PYUPGW_AWS_ID_PROVIDER",
    default="cognito-idp.eu-central-1.amazonaws.com/eu-central-1_HfciXliKM",
)
PYUPGW_AWS_IDENTITY_ENDPOINT = os.getenv(
    "PYUPGW_AWS_IDENTITY_ENDPOINT",
    default="cognito-identity.eu-central-1.amazonaws.com",
)
PYUPGW_AWS_IOT_ENDPOINT = os.getenv(
    "PYUPGW_AWS_IOT_ENDPOINT",
    default="a1b4blxx3o9kj3-ats.iot.eu-central-1.amazonaws.com",
)
PYUPGW_SERVICE_API_BASE_URL = os.getenv(
    "PYUPGW_SERVICE_API_BASE_URL", default="https://service-api.purmo.uleeco.com/api/v1"
)
PYUPGW_SERVICE_API_COMPANY = os.getenv("PYUPGW_SERVICE_API_COMPANY", default="purmo")

_tls_ctx = ClientTlsContext(TlsContextOptions())


class AwsApi:
    """Low level AWS API"""

    def __init__(
        self,
        username: str,
        region: str | None = None,
        client_id: str | None = None,
        user_pool_id: str | None = None,
        id_provider: str | None = None,
        identity_endpoint: str | None = None,
        iot_endpoint: str | None = None,
    ):
        self._region = region or PYUPGW_AWS_REGION
        self._id_provider = id_provider or PYUPGW_AWS_ID_PROVIDER
        self._identity_endpoint = identity_endpoint or PYUPGW_AWS_IDENTITY_ENDPOINT
        self._iot_endpoint = iot_endpoint or PYUPGW_AWS_IOT_ENDPOINT
        self._cognito = Cognito(
            user_pool_id=user_pool_id or PYUPGW_AWS_USER_POOL_ID,
            client_id=client_id or PYUPGW_AWS_CLIENT_ID,
            username=username,
        )
        self._identity_lock = threading.Lock()

    def authenticate(self, password: str) -> tuple[str, str]:
        """Authenticate with ``password``

        Returns:
          Tuple containing id token and access token
        """
        with self._identity_lock:
            self._cognito.authenticate(password)
        return self._cognito.id_token, self._cognito.access_token

    def check_token(self):
        """Check identity token and refresh if necessary

        Returns:
          ``True`` if the token was refreshed, ``False`` otherwise
        """
        with self._identity_lock:
            return self._cognito.check_token(renew=True)

    def get_tokens(self) -> tuple[str, str]:
        """Get identity and access tokens from a previous authentication

        Returns:
          Tuple containing id token and access token
        """
        self.check_token()
        return self._cognito.id_token, self._cognito.access_token

    def get_credentials_provider(self, identity_id: str) -> AwsCredentialsProvider:
        """Get credentials for a previously authenticated identity"""
        return AwsCredentialsProvider.new_cognito(
            endpoint=self._identity_endpoint,
            identity=identity_id,
            logins=[(self._id_provider, self._cognito.id_token)],
            tls_ctx=_tls_ctx,
        )

    async def get_iot_shadow_client(
        self,
        device_code: str,
        credentials_provider: AwsCredentialsProvider,
    ):
        """Get shadow client for a device"""
        mqtt_connection = websockets_with_default_aws_signing(
            endpoint=self._iot_endpoint,
            region=self._region,
            credentials_provider=credentials_provider,
            client_id=f"{device_code}-{uuid.uuid4()}",
            clean_session=False,
            keep_alive_secs=30,
        )
        await asyncio.wrap_future(mqtt_connection.connect())
        return IotShadowClient(mqtt_connection)


class ServiceApi:
    """Low level service API"""

    def __init__(
        self,
        base_url: str | None = None,
        company: str | None = None,
    ):
        self._base_url = base_url or PYUPGW_SERVICE_API_BASE_URL
        self._company = company or PYUPGW_SERVICE_API_COMPANY

    async def _service_api_get(
        self,
        endpoint: str,
        id_token: str,
        access_token: str,
        client_session: aiohttp.ClientSession | None = None,
        **kwargs,
    ):
        headers: dict[str, str] = kwargs.pop("headers", {})
        headers.update(
            {
                "x-auth-token": id_token,
                "x-company-code": self._company,
                "x-access-token": access_token,
            }
        )
        cs_ctx = (
            typing.cast(
                contextlib.AbstractAsyncContextManager[aiohttp.ClientSession],
                contextlib.nullcontext(client_session),
            )
            if client_session
            else aiohttp.ClientSession()
        )
        async with cs_ctx as cs:
            async with cs.get(
                f"{self._base_url}/{endpoint}",
                headers=headers,
                **kwargs,
            ) as response:
                return await response.json()

    def get_slider_list(
        self,
        id_token: str,
        access_token: str,
        client_session: aiohttp.ClientSession | None = None,
    ) -> Awaitable[typing.Any]:
        """Get slider list from the service API"""
        return self._service_api_get(
            "occupants/slider_list", id_token, access_token, client_session
        )

    def get_slider_details(
        self,
        slider_id: str,
        slider_type: str,
        id_token: str,
        access_token: str,
        client_session: aiohttp.ClientSession | None = None,
    ) -> Awaitable[typing.Any]:
        """Get slider details from the service API"""
        return self._service_api_get(
            "occupants/slider_details",
            id_token,
            access_token,
            client_session,
            params={"id": slider_id, "type": slider_type},
        )



"""Main client code"""

import asyncio
import contextlib
import functools
import logging
import typing
import uuid
from collections.abc import Callable, Iterable, Mapping

import aiohttp
from awscrt.mqtt import QoS
from awsiot.iotshadow import (
    ErrorResponse,
    GetShadowRequest,
    GetShadowResponse,
    GetShadowSubscriptionRequest,
    IotShadowClient,
    ShadowState,
    UpdateShadowRequest,
    UpdateShadowResponse,
    UpdateShadowSubscriptionRequest,
)
from dict_deep import deep_get

from ._api import AwsApi, AwsCredentialsProvider, ServiceApi
from .models import (
    Device,
    DeviceType,
    Gateway,
    GatewayAttributes,
    HvacAttributes,
    HvacDevice,
    Occupant,
    SystemMode,
)

if typing.TYPE_CHECKING:
    import concurrent.futures

logger = logging.getLogger(__name__)


class ClientError(Exception):
    """Error in client operation"""


def _parse_device_attributes(data):
    return {
        "id": uuid.UUID(data["id"]),
        "device_code": str(data["device_code"]),
        "model": str(data["model"]),
        "name": str(data["name"]),
    }


def _parse_gateway_attributes(data):
    gateway_data = data["gateway"]
    attributes = _parse_device_attributes(gateway_data)
    occupant_data = gateway_data["occupants_permissions"]["receiver_occupant"]
    occupant = Occupant(
        id=uuid.UUID(occupant_data["id"]),
        identity_id=str(occupant_data["identity_id"]),
    )
    return GatewayAttributes(type=DeviceType.GATEWAY, **attributes, occupant=occupant)


def _parse_hvac_devices(data):
    for item_data in data["items"]:
        if "items" in item_data:
            yield from _parse_hvac_devices(item_data)
        # the gateway itself appears under items, but let's exclude it
        elif "device_code" in item_data and "occupants_permissions" not in item_data:
            yield HvacAttributes(
                type=DeviceType.HVAC, **_parse_device_attributes(item_data)
            )


_SHADOW_TO_ATTRIBUTES_MAP: list[tuple[str, str, Callable[[typing.Any], typing.Any]]] = [
    ("temperature", "ep1:sTherS:HeatingSetpoint_x100", lambda v: float(v) / 100),
    (
        "current_temperature",
        "ep1:sTherS:LocalTemperature_x100",
        lambda v: float(v) / 100,
    ),
    ("min_temp", "ep1:sTherS:MinHeatSetpoint_x100", lambda v: float(v) / 100),
    ("max_temp", "ep1:sTherS:MaxHeatSetpoint_x100", lambda v: float(v) / 100),
    ("system_mode", "ep1:sTherS:RunningMode", SystemMode),
]


def _parse_shadow_attributes(shadow_state: Mapping[str, typing.Any]):
    ret = {}
    for attr_key, shadow_key, transform in _SHADOW_TO_ATTRIBUTES_MAP:
        if (
            value := deep_get(shadow_state, ["11", "properties", shadow_key])
        ) is not None:
            try:
                ret[attr_key] = transform(value)
            except Exception as ex:  # pylint: disable=broad-exception-caught
                logger.warning(
                    "Failed to parse state argument %s=%s",
                    shadow_key,
                    repr(value),
                    exc_info=ex,
                )
    return ret


_ATTRIBUTES_TO_SHADOW_MAP: list[tuple[str, str, Callable[[typing.Any], typing.Any]]] = [
    (
        "temperature",
        "ep1:sTherS:SetHeatingSetpoint_x100",
        lambda v: int(round(v * 100)),
    ),
    ("system_mode", "ep1:sTherS:SetSystemMode", lambda v: v.value),
]


def _create_shadow_update_attributes(changes: Mapping[str, typing.Any]):
    desired_properties: dict[str, typing.Any] = {}
    for attr_key, shadow_key, transform in _ATTRIBUTES_TO_SHADOW_MAP:
        if (value := changes.get(attr_key)) is not None:
            desired_properties[shadow_key] = transform(value)
    return ShadowState(desired={"11": {"properties": desired_properties}})


async def _construct_client_data(id_token: str, access_token: str, client: "Client"):
    service_api = _create_service_api()
    gateways = []
    async with aiohttp.ClientSession() as aiohttp_session:
        slider_list = await service_api.get_slider_list(
            id_token, access_token, aiohttp_session
        )
        logger.debug(
            "Fetched list of gateways: %r",
            slider_list,
            extra={"response": slider_list},
        )
        for gateway_data in slider_list["data"]:
            if gateway_data.get("type") == "gateway":
                attributes = _parse_gateway_attributes(gateway_data)
                slider_details = await service_api.get_slider_details(
                    str(attributes.id),
                    attributes.type.value,
                    id_token,
                    access_token,
                    aiohttp_session,
                )
                logger.debug(
                    "Fetched details for gateway %s: %r",
                    attributes.id,
                    slider_list,
                    extra={"response": slider_list},
                )
                gateways.append(
                    Gateway(
                        attributes,
                        _parse_hvac_devices(slider_details["data"]),
                        client.refresh_device_state,
                        client.update_device_state,
                    )
                )
    return gateways


def _create_aws_api(username: str):
    return AwsApi(username)


def _create_service_api():
    return ServiceApi()


class _CredentialsStore:
    """Cache and rotate AWS credentials"""

    def __init__(self, api: AwsApi):
        self._api = api
        self._credentials_providers: dict[uuid.UUID, AwsCredentialsProvider] = {}
        self._wrapped_credentials_providers: dict[
            uuid.UUID, AwsCredentialsProvider
        ] = {}

    def credentials_provider_for_occupant(self, occupant: Occupant):
        """Get credentials provider for an occupant"""
        credentials_provider = self._credentials_providers.get(occupant.id)
        if not credentials_provider:
            wrapped_credentials_provider = self._api.get_credentials_provider(
                occupant.identity_id
            )
            self._wrapped_credentials_providers[
                occupant.id
            ] = wrapped_credentials_provider
            credentials_provider = self._create_credentials_provider(occupant)
            self._credentials_providers[occupant.id] = credentials_provider
        return credentials_provider

    def _create_credentials_provider(self, occupant: Occupant):
        return AwsCredentialsProvider.new_delegate(
            functools.partial(self._get_credentials, occupant)
        )

    def _get_credentials(self, occupant: Occupant):
        token_expired = self._api.check_token()
        if token_expired:
            wrapped_credentials_provider = self._api.get_credentials_provider(
                occupant.identity_id
            )
            self._wrapped_credentials_providers[
                occupant.id
            ] = wrapped_credentials_provider
        else:
            wrapped_credentials_provider = self._wrapped_credentials_providers[
                occupant.id
            ]
        return wrapped_credentials_provider.get_credentials().result()


class _MqttClientManager(contextlib.AbstractAsyncContextManager):
    """Manage shadow clients for devices"""

    def __init__(self, aws: AwsApi, credentials_store: _CredentialsStore):
        self._aws = aws
        self._credentials_store = credentials_store
        self._on_update_callbacks: list[
            Callable[[str, str, UpdateShadowResponse], None]
        ] = []
        self._shadow_clients: dict[uuid.UUID, IotShadowClient] = {}
        self._pending_responses: dict[str, asyncio.Future] = {}

    async def __aexit__(self, exc_type, exc_value, traceback):
        await asyncio.gather(
            *(
                asyncio.wrap_future(client.mqtt_connection.disconnect())
                for client in self._shadow_clients.values()
            )
        )

    def register_callback(
        self, callback: Callable[[str, str, GetShadowResponse], None]
    ):
        """Register callback that will be invoked when device state is updated

        The arguments to the callback will be the gateway device code, child
        device code, and the response, respectively.
        """
        self._on_update_callbacks.append(callback)

    def remove_callback(self, callback: Callable[[str, str, GetShadowResponse], None]):
        """Remove previously registered callback"""
        self._on_update_callbacks.remove(callback)

    async def client_for_gateway(
        self,
        occupant: Occupant,
        device_code: str,
        child_device_codes: Iterable[str],
    ) -> IotShadowClient:
        """Get shadow client for a given gateway"""
        client = self._shadow_clients.get(occupant.id)
        if not client:
            client = await self._build_shadow_client_with_subscriptions(
                occupant, device_code, child_device_codes
            )
            self._shadow_clients[occupant.id] = client
        return client

    @contextlib.asynccontextmanager
    async def publishing(self):
        """Publishing context that helps waiting for reply"""
        client_token = str(uuid.uuid4())
        exit_stack = contextlib.ExitStack()

        async def wrap_publish(
            publish: Callable[[typing.Any, QoS], "concurrent.futures.Future"],
            request: typing.Any,
        ):
            loop = asyncio.get_event_loop()
            response_future = loop.create_future()
            self._pending_responses[client_token] = response_future
            exit_stack.callback(self._pending_responses.pop, client_token)
            await asyncio.wrap_future(
                publish(
                    request,
                    QoS.AT_MOST_ONCE,
                )
            )
            return await response_future

        with exit_stack:
            yield wrap_publish, client_token

    async def _build_shadow_client_with_subscriptions(
        self,
        occupant: Occupant,
        device_code: str,
        child_device_codes: Iterable[str],
    ):
        loop = asyncio.get_event_loop()
        credentials_provider = await asyncio.to_thread(
            functools.partial(
                self._credentials_store.credentials_provider_for_occupant,
                occupant,
            )
        )
        shadow_client = await self._aws.get_iot_shadow_client(
            device_code, credentials_provider
        )
        subscription_futures = []
        for child_device_code in child_device_codes:
            bound_on_get = functools.partial(
                loop.call_soon_threadsafe,
                self._on_get_callback,
            )
            bound_on_update = functools.partial(
                loop.call_soon_threadsafe,
                self._on_update_callback,
                device_code,
                child_device_code,
            )
            bound_on_error = functools.partial(
                loop.call_soon_threadsafe,
                self._on_error_callback,
            )
            subscription_futures.extend(
                [
                    shadow_client.subscribe_to_get_shadow_accepted(
                        GetShadowSubscriptionRequest(thing_name=child_device_code),
                        QoS.AT_MOST_ONCE,
                        bound_on_get,
                    ),
                    shadow_client.subscribe_to_get_shadow_rejected(
                        GetShadowSubscriptionRequest(thing_name=child_device_code),
                        QoS.AT_MOST_ONCE,
                        bound_on_error,
                    ),
                    shadow_client.subscribe_to_update_shadow_accepted(
                        UpdateShadowSubscriptionRequest(thing_name=child_device_code),
                        QoS.AT_MOST_ONCE,
                        bound_on_update,
                    ),
                    shadow_client.subscribe_to_update_shadow_rejected(
                        UpdateShadowSubscriptionRequest(thing_name=child_device_code),
                        QoS.AT_MOST_ONCE,
                        bound_on_error,
                    ),
                ]
            )
        await asyncio.gather(
            *(asyncio.wrap_future(future) for (future, _) in subscription_futures)
        )
        return shadow_client

    def _resolve_response_future(
        self,
        response: GetShadowResponse | UpdateShadowResponse,
    ):
        if (
            (client_token := response.client_token)
            and (response_future := self._pending_responses.get(client_token))
            and not response_future.done()
        ):
            response_future.set_result(response)

    def _on_error_callback(
        self,
        response: ErrorResponse,
    ):
        if (
            (client_token := response.client_token)
            and (response_future := self._pending_responses.get(client_token))
            and not response_future.done()
        ):
            response_future.set_exception(
                ClientError(f"Request rejected: {response.message} ({response.code})")
            )

    def _on_get_callback(
        self,
        response: GetShadowResponse,
    ):
        self._resolve_response_future(response)

    def _on_update_callback(
        self,
        device_code: str,
        child_device_code: str,
        response: UpdateShadowResponse,
    ):
        self._resolve_response_future(response)
        for callback in self._on_update_callbacks:
            callback(device_code, child_device_code, response)


class Client(contextlib.AbstractAsyncContextManager):
    """Unisenza Plus client

    The recommended way to start a client session is with
    :func:`create_client()` context manager.

    The newly created ``Client`` object doesn't initially know about any devices.
    :meth:`populate_devices()` needs to be called to fetch them from the server.

    Parameters:
      aws: The AWS API object user to access the backend service.  The
           authentication needs to be performed before creting the ``Client``
           object.
    """

    def __init__(self, aws: AwsApi):
        self._exit_stack = contextlib.AsyncExitStack()
        self._aws = aws
        self._gateways: list[Gateway] = []
        self._credentials_store = _CredentialsStore(aws)
        self._mqtt_client_manager = _MqttClientManager(aws, self._credentials_store)
        self._exit_stack.push_async_exit(self._mqtt_client_manager)

    async def aclose(self):
        """Release all resources acquired by the client

        When the ``Client`` object is used as context manager, this is
        automatically called on exit.
        """
        await self._exit_stack.aclose()

    async def __aenter__(self):
        loop = asyncio.get_event_loop()
        callback = functools.partial(loop.call_soon_threadsafe, self._on_update_device)
        self._mqtt_client_manager.register_callback(callback)
        self._exit_stack.callback(self._mqtt_client_manager.remove_callback, callback)
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.aclose()

    async def populate_devices(self):
        """Populate devices from the server"""
        id_token, access_token = await asyncio.to_thread(self._aws.get_tokens)
        self._gateways = await _construct_client_data(id_token, access_token, self)

    def get_gateways(self) -> list[Gateway]:
        """Get the managed gateways"""
        return self._gateways

    def get_devices(self) -> Iterable[tuple[Gateway, HvacDevice]]:
        """Get pairs of gateways and managed devices"""
        for gateway in self._gateways:
            for child in gateway.get_children():
                yield gateway, child

    async def refresh_all_devices(self):
        """Refresh states of all managed devices

        This is a convenience function that calls :meth:`refresh_device_state()`
        for all devices known to the client.

        Raises:
          ClientError: if the request to get device state fails
        """
        publish_futures = []
        for gateway, child in self.get_devices():
            publish_futures.append(self.refresh_device_state(gateway, child))
        await asyncio.gather(*publish_futures)

    async def refresh_device_state(
        self,
        gateway: Gateway,
        device: Device,
    ):
        """Refresh state of a device from the server

        The coroutine completes when the server has responded with the new state.

        Arguments:
          gateway: the gateway the device is connected to
          device: the device whose state is refreshed

        Raises:
          ClientError: if the request to get device state fails
        """
        client = await self._mqtt_client_for_gateway(gateway)
        async with self._mqtt_client_manager.publishing() as (
            wrap_publish,
            client_token,
        ):
            request = GetShadowRequest(
                thing_name=device.get_device_code(), client_token=client_token
            )
            logger.debug(
                "Publishing get shadow request for %s: %r",
                device.get_device_code(),
                request,
                extra={"request": request},
            )
            response = await wrap_publish(client.publish_get_shadow, request)
            logger.debug(
                "Get shadow response for %s: %r",
                device.get_device_code(),
                response,
                extra={"response": response},
            )
        changes = _parse_shadow_attributes(response.state.reported)
        device.set_attributes(changes)

    async def update_device_state(
        self,
        gateway: Gateway,
        device: Device,
        changes: Mapping[str, typing.Any],
    ):
        """Update the state of a device managed by the client

        This method will send the changed values to the server.  The coroutine
        will complete after the server has accepted the request, but not yet
        necessarily applied the changes.

        The update itself is asynchronous, and the in-memory attributes of the
        device models will only be updated after the server has acknowledged
        that it has applied the changes.  :meth:`Device.subscribe()` can be used
        to subscribe to the updates of the in-memory model.

        Arguments:
          gateway: the gateway the device is connected to
          device: the device whose state is updated
          changes: the changes to be published

        Raises:
          ClientError: if the request to update device state fails
        """
        client = await self._mqtt_client_for_gateway(gateway)
        async with self._mqtt_client_manager.publishing() as (
            wrap_publish,
            client_token,
        ):
            request = UpdateShadowRequest(
                client_token=client_token,
                thing_name=device.get_device_code(),
                state=_create_shadow_update_attributes(changes),
            )
            logger.debug(
                "Publishing update shadow request for %s: %r",
                device.get_device_code(),
                request,
                extra={"request": request},
            )
            await wrap_publish(client.publish_update_shadow, request)

    def _mqtt_client_for_gateway(self, gateway: Gateway):
        return self._mqtt_client_manager.client_for_gateway(
            gateway.get_occupant(),
            gateway.get_device_code(),
            [child.get_device_code() for child in gateway.get_children()],
        )

    def _on_update_device(
        self,
        device_code: str,
        child_device_code: str,
        response: GetShadowResponse | UpdateShadowResponse,
    ):
        logger.debug(
            "Received update for %s in %s: %r",
            child_device_code,
            device_code,
            response,
            extra={"response": response},
        )
        for gateway in self._gateways:
            if gateway.get_device_code() == device_code:
                for child in gateway.get_children():
                    if child.get_device_code() == child_device_code:
                        changes = _parse_shadow_attributes(response.state.reported)
                        child.set_attributes(changes)


@contextlib.asynccontextmanager
async def create_client(username: str, password: str):
    """Create Unisenza Plus client

    This function returns a context manager that initializes and manages
    resources for a :class:`Client` instance.  It also takes care of
    authenticating and populating devices from the server.

    .. code-block:: python

        async with create_client("user@example.com", "password") as client:
           ...  # use client

    Arguments:
      username, password: the credentials used to log into the cloud service
    """

    aws = _create_aws_api(username)
    logger.debug("Authenticating user %s", username)
    await asyncio.to_thread(functools.partial(aws.authenticate, password))
    async with Client(aws) as client:
        await client.populate_devices()
        yield client



"""Tests for the UPGW client"""

# pylint: disable=redefined-outer-name

import contextlib
import unittest.mock
from concurrent.futures import Future

import pytest
from attrs import define
from hypothesis import given
from hypothesis import strategies as st

import pyupgw.client
from pyupgw import GatewayAttributes, HvacAttributes, create_client

ID_TOKEN = "id_token"
ACCESS_TOKEN = "access_token"
USERNAME = "username"
PASSWORD = "password"


GatewayData = tuple[GatewayAttributes, list[HvacAttributes]]


@define
class _MockAws:
    authenticate: unittest.mock.Mock
    check_token: unittest.mock.Mock
    get_tokens: unittest.mock.Mock
    get_credentials_provider: unittest.mock.Mock
    get_iot_shadow_client: unittest.mock.AsyncMock


@define
class _MockShadowClient:
    subscribe_to_get_shadow_accepted: unittest.mock.Mock
    subscribe_to_update_shadow_accepted: unittest.mock.Mock
    publish_get_shadow: unittest.mock.Mock
    publish_update_shadow: unittest.mock.Mock
    mqtt_connection: unittest.mock.MagicMock


def _instant_future(result):
    future = Future()
    future.set_result(result)
    return future


def _mock_aws(monkeypatch) -> tuple[_MockAws, _MockShadowClient]:
    mock_shadow_client = _MockShadowClient(
        subscribe_to_get_shadow_accepted=unittest.mock.Mock(
            side_effect=lambda *_, **__: (_instant_future(None), 0)
        ),
        subscribe_to_update_shadow_accepted=unittest.mock.Mock(
            side_effect=lambda *_, **__: (_instant_future(None), 0)
        ),
        publish_get_shadow=unittest.mock.Mock(
            side_effect=lambda *_, **__: _instant_future(None)
        ),
        publish_update_shadow=unittest.mock.Mock(
            side_effect=lambda *_, **__: _instant_future(None)
        ),
        mqtt_connection=unittest.mock.MagicMock(),
    )
    mock_shadow_client.mqtt_connection.disconnect.side_effect = lambda: _instant_future(
        None
    )
    mock_aws = _MockAws(
        authenticate=unittest.mock.Mock(return_value=(ID_TOKEN, ACCESS_TOKEN)),
        check_token=unittest.mock.Mock(return_value=False),
        get_tokens=unittest.mock.Mock(return_value=(ID_TOKEN, ACCESS_TOKEN)),
        get_credentials_provider=unittest.mock.Mock(return_value=object()),
        get_iot_shadow_client=unittest.mock.AsyncMock(return_value=mock_shadow_client),
    )
    monkeypatch.setattr(pyupgw.client, "_create_aws_api", lambda username: mock_aws)
    return mock_aws, mock_shadow_client


@define
class _MockServiceApi:
    get_slider_list: unittest.mock.AsyncMock
    get_slider_details: unittest.mock.AsyncMock


def _mock_service_api(monkeypatch, gateways: list[GatewayData]) -> _MockServiceApi:
    slider_list_data = [
        {
            "id": str(attributes.id),
            "type": attributes.type.value,
            "gateway": {
                "id": str(attributes.id),
                "device_code": attributes.device_code,
                "model": attributes.model,
                "name": attributes.name,
                "occupants_permissions": {
                    "receiver_occupant": {
                        "id": str(attributes.occupant.id),
                        "identity_id": attributes.occupant.identity_id,
                    }
                },
            },
        }
        for (attributes, _) in gateways
    ]

    slider_details_map = {
        slider_data["id"]: {
            "data": {
                **slider_data,
                "items": [
                    slider_data["gateway"],
                    *(
                        {
                            "id": str(item.id),
                            "device_code": item.device_code,
                            "model": item.model,
                            "name": item.name,
                        }
                        for item in gateways[i][1]
                    ),
                ],
            }
        }
        for (i, slider_data) in enumerate(slider_list_data)
    }

    def get_slider_details_impl(slider_id, *_, **__):
        return slider_details_map[str(slider_id)]

    ret = _MockServiceApi(
        get_slider_list=unittest.mock.AsyncMock(
            return_value={"data": slider_list_data}
        ),
        get_slider_details=unittest.mock.AsyncMock(side_effect=get_slider_details_impl),
    )
    monkeypatch.setattr(pyupgw.client, "_create_service_api", lambda: ret)

    return ret


@pytest.fixture(scope="session")
def client_setup():
    """Context manager to setup low level clients"""

    @contextlib.contextmanager
    def context(gateways: list[GatewayData]):
        with pytest.MonkeyPatch().context() as m:
            aws, shadow_client = _mock_aws(m)
            service_api = _mock_service_api(m, gateways)
            yield aws, service_api, shadow_client

    return context


@pytest.mark.asyncio
@given(
    gateways=st.lists(
        st.tuples(
            st.builds(GatewayAttributes),
            st.lists(
                st.builds(
                    HvacAttributes,
                    system_mode=st.none(),
                    temperature=st.none(),
                    current_temperature=st.none(),
                    min_temp=st.none(),
                    max_temp=st.none(),
                )
            ),
        )
    )
)
async def test_get_gateways(gateways: list[GatewayData], client_setup):
    with client_setup(gateways) as (aws, service_api, _):
        async with create_client(USERNAME, PASSWORD) as client:
            for (
                expected_attributes,
                expected_children_attributes,
            ), actual_gateway in zip(gateways, client.get_gateways()):
                assert actual_gateway.get_attributes() == expected_attributes
                assert [
                    device.get_attributes() for device in actual_gateway.get_children()
                ] == expected_children_attributes

        aws.authenticate.assert_called_once()
        service_api.get_slider_list.assert_awaited_once_with(
            ID_TOKEN, ACCESS_TOKEN, unittest.mock.ANY
        )
        for expected_attributes, _ in gateways:
            service_api.get_slider_details.assert_any_await(
                str(expected_attributes.id),
                expected_attributes.type.value,
                ID_TOKEN,
                ACCESS_TOKEN,
                unittest.mock.ANY,
            )



"""Tests for data models"""

import logging
import unittest.mock

import pytest
from attrs import asdict
from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st

from pyupgw import Gateway, GatewayAttributes, HvacAttributes, HvacDevice, SystemMode


@given(
    st.builds(
        HvacAttributes,
        temperature=st.floats(allow_nan=False),
        current_temperature=st.floats(allow_nan=False),
        min_temp=st.floats(allow_nan=False),
        max_temp=st.floats(allow_nan=False),
    )
)
def test_device(attributes: HvacAttributes):
    device = HvacDevice(
        attributes, unittest.mock.AsyncMock(), unittest.mock.AsyncMock()
    )
    assert device.get_attributes() == attributes
    assert device.get_id() == attributes.id
    assert device.get_type() == attributes.type
    assert device.get_device_code() == attributes.device_code
    assert device.get_model() == attributes.model
    assert device.get_name() == attributes.name
    assert device.get_system_mode() == attributes.system_mode
    assert device.get_temperature() == attributes.temperature
    assert device.get_current_temperature() == attributes.current_temperature
    assert device.get_min_temp() == attributes.min_temp
    assert device.get_max_temp() == attributes.max_temp


@given(...)
def test_device_set_attributes(
    attributes: HvacAttributes, new_attributes: HvacAttributes
):
    subscriber = unittest.mock.Mock()
    device = HvacDevice(
        attributes, unittest.mock.AsyncMock(), unittest.mock.AsyncMock()
    )
    device.subscribe(subscriber)
    device.set_attributes(asdict(new_attributes))
    assert device.get_attributes() == new_attributes
    subscriber.assert_called_with(device, asdict(new_attributes))


@given(...)
def test_device_set_attributes_empty_change(attributes: HvacAttributes):
    subscriber = unittest.mock.Mock()
    device = HvacDevice(
        attributes, unittest.mock.AsyncMock(), unittest.mock.AsyncMock()
    )
    device.subscribe(subscriber)
    device.set_attributes({})
    assert device.get_attributes() == attributes
    subscriber.assert_not_called()


@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(attributes=..., new_attributes=...)
def test_device_set_attributes_callback_error(
    attributes: HvacAttributes, new_attributes: HvacAttributes, caplog
):
    subscriber = unittest.mock.Mock(side_effect=Exception("bad error"))
    device = HvacDevice(
        attributes, unittest.mock.AsyncMock(), unittest.mock.AsyncMock()
    )
    device.subscribe(subscriber)
    with caplog.at_level(logging.ERROR):
        device.set_attributes(asdict(new_attributes))
    caplog.clear()
    assert device.get_attributes() == new_attributes
    subscriber.assert_called_with(device, asdict(new_attributes))


@pytest.mark.asyncio
@given(...)
async def test_device_refresh(attributes: HvacAttributes):
    dispatch_refresh = unittest.mock.AsyncMock()
    device = HvacDevice(attributes, dispatch_refresh, unittest.mock.AsyncMock())
    await device.refresh()
    dispatch_refresh.assert_awaited_with(device)


@pytest.mark.asyncio
@given(...)
async def test_device_update(
    attributes: HvacAttributes, new_attributes: HvacAttributes
):
    dispatch_update = unittest.mock.AsyncMock()
    device = HvacDevice(attributes, unittest.mock.AsyncMock(), dispatch_update)
    await device.update(asdict(new_attributes))
    dispatch_update.assert_awaited_with(device, asdict(new_attributes))


@pytest.mark.asyncio
@given(...)
async def test_device_update_system_mode(
    attributes: HvacAttributes, system_mode: SystemMode
):
    dispatch_update = unittest.mock.AsyncMock()
    device = HvacDevice(attributes, unittest.mock.AsyncMock(), dispatch_update)
    await device.update_system_mode(system_mode)
    dispatch_update.assert_awaited_with(device, {"system_mode": system_mode})


@pytest.mark.asyncio
@given(attributes=..., temperature=st.floats(allow_nan=False))
async def test_device_update_temperature(
    attributes: HvacAttributes, temperature: float
):
    dispatch_update = unittest.mock.AsyncMock()
    device = HvacDevice(attributes, unittest.mock.AsyncMock(), dispatch_update)
    await device.update_temperature(temperature)
    dispatch_update.assert_awaited_with(device, {"temperature": temperature})


@given(...)
def test_gateway(attributes: GatewayAttributes, children: list[HvacAttributes]):
    gateway = Gateway(
        attributes, children, unittest.mock.AsyncMock(), unittest.mock.AsyncMock()
    )
    assert gateway.get_attributes() == attributes
    assert gateway.get_occupant() == attributes.occupant
    assert [child.get_attributes() for child in gateway.get_children()] == children


@pytest.mark.asyncio
@given(...)
async def test_gateway_refresh_child(
    attributes: GatewayAttributes, child_attributes: HvacAttributes
):
    dispatch_refresh = unittest.mock.AsyncMock()
    gateway = Gateway(
        attributes, [child_attributes], dispatch_refresh, unittest.mock.AsyncMock()
    )
    (child,) = gateway.get_children()
    await child.refresh()
    dispatch_refresh.assert_awaited_with(gateway, child)


@pytest.mark.asyncio
@given(...)
async def test_gateway_update_child(
    attributes: GatewayAttributes,
    child_attributes: HvacAttributes,
    new_child_attributes: HvacAttributes,
):
    dispatch_update = unittest.mock.AsyncMock()
    gateway = Gateway(
        attributes, [child_attributes], unittest.mock.AsyncMock(), dispatch_update
    )
    (child,) = gateway.get_children()
    await child.update(asdict(new_child_attributes))
    dispatch_update.assert_awaited_with(gateway, child, asdict(new_child_attributes))
